del ..\..\qemu\qapi-types.h 2> null
del ..\..\qemu\qapi-types.c 2> null

del ..\..\qemu\qapi-visit.h 2> null
del ..\..\qemu\qapi-visit.c 2> null

del ..\..\qemu\config-host.h 2> null

del ..\..\qemu\aarch-softmmu\config-target.h 2> null
del ..\..\qemu\aarcheb-softmmu\config-target.h 2> null
del ..\..\qemu\arm-softmmu\config-target.h 2> null
del ..\..\qemu\armeb-softmmu\config-target.h 2> null
del ..\..\qemu\m68k-softmmu\config-target.h 2> null
del ..\..\qemu\mips64el-softmmu\config-target.h 2> null
del ..\..\qemu\mips64-softmmu\config-target.h 2> null
del ..\..\qemu\mipsel-softmmu\config-target.h 2> null
del ..\..\qemu\mips-softmmu\config-target.h 2> null
del ..\..\qemu\sparc64-softmmu\config-target.h 2> null
del ..\..\qemu\sparc-softmmu\config-target.h 2> null
del ..\..\qemu\x86_64-softmmu\config-target.h 2> null
del null
