/* Automatically generated by create_config - do not modify */
#define TARGET_ABI_MIPSN64 1
#define TARGET_MIPS64 1
#define TARGET_NAME "mips64"
#define TARGET_MIPS 1
#define TARGET_WORDS_BIGENDIAN 1
#define CONFIG_SOFTMMU 1
